Option Compare Binary
Option Explicit On
Option Strict On

Imports System.IO
Imports System.Text
Imports System.Text.RegularExpressions

Public Class LibFile
    Private m_FirstLinkerMemberHeader As ArchiveMemberHeader
    Private m_SecondLinkerMemberHeader As ArchiveMemberHeader
    Private m_LongnameMemberHeader As ArchiveMemberHeader
    Private m_ImportSymbols As New List(Of ImportSymbol)()

    ' Contains the longnames with offset as key and name as value.
    Private m_Longnames As New Dictionary(Of Long, String)()

    Public Event ParseError(ByVal sender As Object, ByVal e As ParseErrorEventArgs)

    Protected Sub OnParseError(ByVal e As ParseErrorEventArgs)
        RaiseEvent ParseError(Me, e)
    End Sub

    Public Sub Load(ByVal FileName As String)
        Dim Reader As New ExtendedBinaryReader( _
            New FileStream(FileName, System.IO.FileMode.Open), _
            Encoding.ASCII _
        )
        Try

            ' *** File signature.
            ReadSignature(Reader)

            ' *** First linker archive member.
            Reader.SeekNextEvenPosition()
            Dim FirstLinkerMemberBasePosition As Long = Reader.BaseStream.Position
            Me.FirstLinkerMemberHeader = ReadArchiveMemberHeader(Reader)

            ' Skip archive member offsets and symbol names in the first linker
            ' archive member.
            'Dim MembersCount As Integer = CInt(Reader.ReadBigEndianUInt32())
            'Dim ArchiveMemberHeaderOffsets(MembersCount - 1) As Long
            'For i As Integer = 0 To MembersCount - 1
            '    ArchiveMemberHeaderOffsets(i) = Reader.ReadBigEndianUInt32()
            'Next i
            'For i As Integer = 0 To MembersCount - 1
            '    Dim SymbolName As String = Reader.ReadNullTerminatedString()
            'Next i
            Const HeaderSize As Integer = 60
            Reader.BaseStream.Position = _
                FirstLinkerMemberBasePosition + _
                HeaderSize + _
                Me.FirstLinkerMemberHeader.MemberSize

            ' *** Second linker archive member.
            Reader.SeekNextEvenPosition()
            Me.SecondLinkerMemberHeader = ReadArchiveMemberHeader(Reader)
            Dim MembersCount As Integer = CInt(Reader.ReadUInt32())
            Dim ArchiveMemberHeaderOffsets(MembersCount - 1) As Long
            For i As Integer = 0 To MembersCount - 1
                ArchiveMemberHeaderOffsets(i) = Reader.ReadUInt32()
            Next i
            Dim IndexedSymbolsCount As Integer = CInt(Reader.ReadUInt32())

            ' Store pairs of offsets and archive member entry names in the list.
            Dim Indices(IndexedSymbolsCount - 1) As UInt16
            Dim ImportMemberEntries As New List(Of ImportMemberEntry)
            For i As Integer = 0 To IndexedSymbolsCount - 1
                Indices(i) = Reader.ReadUInt16()
                Dim ex As New ImportMemberEntry()
                ImportMemberEntries.Add(ex)
            Next i
            For i As Integer = 0 To IndexedSymbolsCount - 1
                ImportMemberEntries(i).Name = Reader.ReadNullTerminatedString()
                ImportMemberEntries(i).Offset = ArchiveMemberHeaderOffsets(Indices(i) - 1)
            Next i

            ' *** Long name archive member (must be present but can be empty).
            Reader.SeekNextEvenPosition()
            LongnameMemberHeader = ReadArchiveMemberHeader(Reader)
            Dim HeaderBasePosition As Long = Reader.BaseStream.Position
            Dim BytesRead As Long = Reader.BaseStream.Position - HeaderBasePosition
            Do While BytesRead < LongnameMemberHeader.MemberSize
                m_Longnames.Add(BytesRead, Reader.ReadNullTerminatedString())
                BytesRead = Reader.BaseStream.Position - HeaderBasePosition
            Loop

            ' *** Archive members containing import symbols.
            ' TODO:  Reading currently only works with import symbol descriptions
            ' which are in short format.
            For i As Integer = 0 To IndexedSymbolsCount - 1
                Reader.BaseStream.Seek(ImportMemberEntries(i).Offset, SeekOrigin.Begin)
                Reader.SeekNextEvenPosition()
                ReadArchiveMemberHeader(Reader)
                ImportSymbols.Add(ReadArchiveImportSymbolEntry(Reader))
            Next i
        Catch
            Throw
        Finally
            Reader.Close()
        End Try
    End Sub

    Public Property FirstLinkerMemberHeader() As ArchiveMemberHeader
        Get
            Return m_FirstLinkerMemberHeader
        End Get
        Protected Set(ByVal Value As ArchiveMemberHeader)
            m_FirstLinkerMemberHeader = Value
        End Set
    End Property

    Public Property SecondLinkerMemberHeader() As ArchiveMemberHeader
        Get
            Return m_SecondLinkerMemberHeader
        End Get
        Protected Set(ByVal Value As ArchiveMemberHeader)
            m_SecondLinkerMemberHeader = Value
        End Set
    End Property

    Public Property LongnameMemberHeader() As ArchiveMemberHeader
        Get
            Return m_LongnameMemberHeader
        End Get
        Protected Set(ByVal Value As ArchiveMemberHeader)
            m_LongnameMemberHeader = Value
        End Set
    End Property

    Public ReadOnly Property ImportSymbols() As List(Of ImportSymbol)
        Get
            Return m_ImportSymbols
        End Get
    End Property

    Private Sub ReadSignature(ByVal Reader As ExtendedBinaryReader)
        Const LibSignature As String = "!<arch>" & ControlChars.Lf
        Dim Position As Long = Reader.BaseStream.Position
        If Reader.ReadFixedLengthString(8) <> LibSignature Then
            Me.OnParseError(New ParseErrorEventArgs(Position, "The file is not a valid library file because the signature is invalid."))
        End If
    End Sub

    Private Function ReadArchiveMemberHeader(ByVal Reader As ExtendedBinaryReader) As ArchiveMemberHeader 'Integer
        Dim Header As New ArchiveMemberHeader()

        ' Archive member name.
        ' "<name>/" is the name of a member.
        ' "\<offset>" means that the name of the member is stored in the longname
        '     member.  <offset> is the offset based on the end of the longname
        '     members header's end.
        ' "/" is the name of the first/second linker member.
        ' "//" is the name of the longname member.
        Dim Name As String = RTrim(Reader.ReadFixedLengthString(16))
        If Regex.IsMatch(Name, "[^/]+/") Then
            Name = Strings.Left(Name, Len(Name) - 1)
        ElseIf Regex.IsMatch(Name, "/\d") Then

            ' Obtain name from the longname member.
            Name = m_Longnames(CInt(Mid(Name, 2)))
        End If
        Header.Name = Name

        ' Date and time when the archive member was created.
        Header.Date = #1/1/1970#.ToUniversalTime().AddSeconds(CDbl(Reader.ReadFixedLengthString(12)))

        ' User ID, no meaningful value on Windows, all blanks.
        Header.UserId = RTrim(Reader.ReadFixedLengthString(6))

        ' Group ID, no meaningful value on Windows, all blanks.
        Header.GroupId = RTrim(Reader.ReadFixedLengthString(6))

        ' The member's file mode.
        Header.FileMode = CType(CInt(Reader.ReadFixedLengthString(8)), FileMode)

        ' Total size of the archive member not including the size of its header.
        Header.MemberSize = CInt(Reader.ReadFixedLengthString(10))

        ' The header must end with the character sequence "\n".
        Dim Position As Long = Reader.BaseStream.Position
        If Reader.ReadFixedLengthString(2) <> Chr(&H60) & Chr(&HA) Then
            Me.OnParseError(New ParseErrorEventArgs(Position, "The header end is invalid."))
        End If

        Return Header
    End Function

    ' Read import header.
    Private Function ReadArchiveImportSymbolEntry(ByVal Reader As ExtendedBinaryReader) As ImportSymbol
        Dim Symbol As New ImportSymbol()

        ' Signature 1.
        Dim Position As Long = Reader.BaseStream.Position
        Dim Sig1 As Int16 = Reader.ReadInt16()
        If Sig1 <> IMAGE_FILE_MACHINE.IMAGE_FILE_MACHINE_UNKNOWN Then
            Me.OnParseError(New ParseErrorEventArgs(Position, "Sig1 is not equal to 'IMAGE_FILE_MACHINE_UNKNOWN' (0x0)."))
        End If

        ' Signature 2.
        Position = Reader.BaseStream.Position
        Dim Sig2 As Int16 = Reader.ReadInt16()
        If Sig2 <> &HFFFF Then
            Me.OnParseError(New ParseErrorEventArgs(Position, "Sig2 is not equal to 0xFFFF."))
        End If

        ' Version (signed or unsigned?).
        Dim Version As Int16 = Reader.ReadInt16()

        ' Machine.
        Symbol.Machine = CType(Reader.ReadInt16(), ImageFileMachine)

        ' Timestamp.
        Dim TimeDateStamp As Int32 = Reader.ReadInt32()

        ' Size of data(signed or unsigned?).
        Dim SizeOfData As Int32 = Reader.ReadInt32()

        ' Ordinal/hint (Signed or unsigned?).
        Symbol.OrdinalHint = Reader.ReadInt16()

        ' Import name and type.
        Position = Reader.BaseStream.Position
        Dim Data() As Byte = Reader.ReadBytes(2)    ' 2 + 3 + 11 bits.

        ' Import name type.
        Const ImportNameTypeMask As Byte = 56   ' 00111000.
        Dim ImportNameType As Byte = Data(0) And ImportNameTypeMask
        ImportNameType = ImportNameType >> 3
        Symbol.ImportNameType = CType(ImportNameType, ImportNameType)

        ' Import type.
        Const ImportMask As Byte = 192  ' 11000000.
        Dim ImportType As Byte = Data(0) And ImportMask
        ImportType = ImportType >> 6
        Symbol.ImportType = CType(ImportType, ImportType)

        If Data(0) << 5 <> 0 AndAlso Data(1) <> 0 Then
            Me.OnParseError(New ParseErrorEventArgs(Position, "Reserved 11 bits directly after import name type are not set to 0."))
        End If

        ' Symbol name and DLL name.
        Symbol.Name = Reader.ReadNullTerminatedString()
        Symbol.DllName = Reader.ReadNullTerminatedString()

        Return Symbol
    End Function

    ' This would be possible by using the 'UnDecorateSymbolName' ("Dbghelp.dll") function.
    Public Shared Function GetUndecoratedName(ByVal Name As String) As String
        If Name Is Nothing Then
            Throw New ArgumentException("The value of the 'Name' parameter must not be 'Nothing'.")
        End If
        Select Case Name.Chars(0)
            Case "?"c, "@"c, "_"c
                Name = Mid(Name, 2)
        End Select
        Dim Pos As Integer = InStr(Name, "@")
        If Pos > 0 Then
            Name = Left(Name, Pos - 1)
        End If
        Return Name
    End Function

    Public Shared Function GetDecoratedName(ByVal Name As String, ByVal ParamsLength As Integer) As String
        If Name Is Nothing Then
            Throw New ArgumentException("The value of the 'Name' parameter must not be 'Nothing'.")
        End If
        Return "_" & Name & "@" & ParamsLength
    End Function

    Public Shared Function GetParamsLength(ByVal Name As String) As Integer
        If Name Is Nothing Then
            Throw New ArgumentException("The value of the 'Name' parameter must not be 'Nothing'.")
        End If
        Dim Pos As Integer = InStrRev(Name, "@")
        If Pos > 0 Then
            Dim Value As Integer
            If Integer.TryParse(Mid(Name, Pos + 1), Value) Then
                Return Value
            Else
                Return -1
            End If
        Else
            Return -1
        End If
    End Function
End Class